version = "unknown"

try:
    from pkg_resources import get_distribution

    version = get_distribution(__name__).version
except ImportError:
    try:
        from importlib import metadata

        version = metadata.version("pycoin")
    except Exception:
        pass
except Exception:
    pass

__title__ = "pycoin"
__author__ = "Richard Kiss"
__version__ = version
__license__ = "MIT"
__copyright__ = "Copyright 2018 Richard Kiss"

"""
:copyright: (c) 2018 by Richard Kiss
:license: MIT, see LICENSE for more details.
"""



import unittest

from pycoin.contrib.bech32m import bech32_decode, bech32_encode, decode, encode, Encoding
from pycoin.encoding.hexbytes import b2h


class Bech32Test(unittest.TestCase):
    def test_bip350_vectors(self):
        # from
        # https://github.com/bitcoin/bips/blob/master/bip-0350.mediawiki
        # #Test_vectors_for_v0v16_native_segregated_witness_addresses

        VALID = """
            A1LQFN3A
            a1lqfn3a
            an83characterlonghumanreadablepartthatcontainsthetheexcludedcharactersbioandnumber11sg7hg6
            abcdef1l7aum6echk45nj3s0wdvt2fg8x9yrzpqzd3ryx
            11llllllllllllllllllllllllllllllllllllllllllllllllllllllllllllllllllllllllllllllllllludsr8
            split1checkupstagehandshakeupstreamerranterredcaperredlc445v
            ?1v759aa""".split()
        for _ in VALID:
            hrp, data, spec = bech32_decode(_)
            self.assertEqual(spec, Encoding.BECH32M)
            encoded = bech32_encode(hrp, data, spec)
            self.assertEqual(_.lower(), encoded)

        INVALID = [
            "\x201xj0phk", # : HRP character out of range
            "\x7f1g6xzxy", # : HRP character out of range
            "\x801vctc34", # : HRP character out of range
            "an84characterslonghumanreadablepartthatcontainsthetheexcludedcharactersbioandnumber11d6pts4", # : overall max length exceeded
            "qyrz8wqd2c9m", # : No separator character
            "1qyrz8wqd2c9m", # : Empty HRP
            "y1b0jsk6g", # : Invalid data character
            "lt1igcx5c0", # : Invalid data character
            "in1muywd", # : Too short checksum
            "mm1crxm3i", # : Invalid character in checksum
            "au1s5cgom", # : Invalid character in checksum
            "M1VUXWEZ", # : checksum calculated with uppercase form of HRP
            "16plkw9", # : empty HRP
            "1p2gdwpf", # : e
        ]
        for _ in INVALID:
            hrp, data, spec = bech32_decode(_)
            self.assertIsNone(hrp)
            self.assertIsNone(data)
            self.assertIsNone(spec)

        ADDRESSES = """BC1QW508D6QEJXTDG4Y5R3ZARVARY0C5XW7KV8F3T4: 0014751e76e8199196d454941c45d1b3a323f1433bd6
            tb1qrp33g0q5c5txsp9arysrx4k6zdkfs4nce4xj0gdcccefvpysxf3q0sl5k7: 00201863143c14c5166804bd19203356da136c985678cd4d27a1b8c6329604903262
            bc1pw508d6qejxtdg4y5r3zarvary0c5xw7kw508d6qejxtdg4y5r3zarvary0c5xw7kt5nd6y: 5128751e76e8199196d454941c45d1b3a323f1433bd6751e76e8199196d454941c45d1b3a323f1433bd6
            BC1SW50QGDZ25J: 6002751e
            bc1zw508d6qejxtdg4y5r3zarvaryvaxxpcs: 5210751e76e8199196d454941c45d1b3a323
            tb1qqqqqp399et2xygdj5xreqhjjvcmzhxw4aywxecjdzew6hylgvsesrxh6hy: 0020000000c4a5cad46221b2a187905e5266362b99d5e91c6ce24d165dab93e86433
            tb1pqqqqp399et2xygdj5xreqhjjvcmzhxw4aywxecjdzew6hylgvsesf3hn0c: 5120000000c4a5cad46221b2a187905e5266362b99d5e91c6ce24d165dab93e86433
            bc1p0xlxvlhemja6c4dqv22uapctqupfhlxm9h8z3k2e72q4k9hcz7vqzk5jj0: 512079be667ef9dcbbac55a06295ce870b07029bfcdb2dce28d959f2815b16f81798"""

        for line in ADDRESSES.split("\n"):
            address, script_pub_key_hex = line.split(": ")
            address = address.strip()
            pos = address.rfind("1")
            hrp = address[:pos].lower()
            v = decode(hrp, address)
            version, blob = v
            if version != 0:
                version += 0x50
            calc_hex = b2h(bytearray([version] + [len(blob)] + blob))
            self.assertEqual(calc_hex, script_pub_key_hex)

        BAD_ADDRESSES = """tc1p0xlxvlhemja6c4dqv22uapctqupfhlxm9h8z3k2e72q4k9hcz7vq5zuyut: Invalid human-readable part
            bc1p0xlxvlhemja6c4dqv22uapctqupfhlxm9h8z3k2e72q4k9hcz7vqh2y7hd: Invalid checksum (Bech32 instead of Bech32m)
            tb1z0xlxvlhemja6c4dqv22uapctqupfhlxm9h8z3k2e72q4k9hcz7vqglt7rf: Invalid checksum (Bech32 instead of Bech32m)
            BC1S0XLXVLHEMJA6C4DQV22UAPCTQUPFHLXM9H8Z3K2E72Q4K9HCZ7VQ54WELL: Invalid checksum (Bech32 instead of Bech32m)
            bc1qw508d6qejxtdg4y5r3zarvary0c5xw7kemeawh: Invalid checksum (Bech32m instead of Bech32)
            tb1q0xlxvlhemja6c4dqv22uapctqupfhlxm9h8z3k2e72q4k9hcz7vq24jc47: Invalid checksum (Bech32m instead of Bech32)
            bc1p38j9r5y49hruaue7wxjce0updqjuyyx0kh56v8s25huc6995vvpql3jow4: Invalid character in checksum
            BC130XLXVLHEMJA6C4DQV22UAPCTQUPFHLXM9H8Z3K2E72Q4K9HCZ7VQ7ZWS8R: Invalid witness version
            bc1pw5dgrnzv: Invalid program length (1 byte)
            bc1p0xlxvlhemja6c4dqv22uapctqupfhlxm9h8z3k2e72q4k9hcz7v8n0nx0muaewav253zgeav: Invalid program length (41 bytes)
            BC1QR508D6QEJXTDG4Y5R3ZARVARYV98GJ9P: Invalid program length for witness version 0 (per BIP141)
            tb1p0xlxvlhemja6c4dqv22uapctqupfhlxm9h8z3k2e72q4k9hcz7vq47Zagq: Mixed case
            bc1p0xlxvlhemja6c4dqv22uapctqupfhlxm9h8z3k2e72q4k9hcz7v07qwwzcrf: zero padding of more than 4 bits
            tb1p0xlxvlhemja6c4dqv22uapctqupfhlxm9h8z3k2e72q4k9hcz7vpggkg4j: Non-zero padding in 8-to-5 conversion
            bc1gmk9yu"""

        for line in BAD_ADDRESSES.split("\n"):
            address = line.split(": ")[0].strip()
            pos = address.rfind("1")
            hrp = address[:pos].lower()
            if hrp not in ["tb", "bc"]:
                continue
            v = decode(hrp, address)
            self.assertEqual(v, (None, None))



from pycoin.networks.bitcoinish import create_bitcoinish_network

network = create_bitcoinish_network(
    symbol="MZC", network_name="Mazacoin", subnet_name="mainnet",
    wif_prefix_hex="e0", address_prefix_hex="32", pay_to_script_prefix_hex="5c39",
    bip32_prv_prefix_hex="0488ade4", bip32_pub_prefix_hex="0488b21e")



import hashlib
import os

import pycoin.contrib.ripemd160

from .hexbytes import bytes_as_revhex


def ripemd160_native(data):
    return hashlib.new("ripemd160", data)


def get_best_ripemd160():
    # ubuntu 22 features an openssl without ripemd160, where python gets its
    # implementation from. To top it off, `"ripemd160" in hashlib.algorithms_available`
    # still evaluates to true, so we actually have to try it to see if we'll fail.

    USE_NATIVE = "ripemd160" in hashlib.algorithms_available and not os.getenv(
        "PYCOIN_USE_PYTHON_RIPEMD160"
    )

    if USE_NATIVE:
        try:
            ripemd160_native(b"").digest()
            return ripemd160_native
        except Exception:
            pass

    # stupid Google App Engine hashlib doesn't support ripemd160 for some stupid reason
    # import it from pycrypto. You need to add
    # - name: pycrypto
    #   version: "latest"
    # to the "libraries" section of your app.yaml
    try:
        from Crypto.Hash.RIPEMD import RIPEMD160Hash
    except Exception:

        class RIPEMD160Hash:
            def __init__(self, data):
                self._digest = pycoin.contrib.ripemd160.ripemd160(data)

            def digest(self):
                return self._digest

    return RIPEMD160Hash


ripemd160 = get_best_ripemd160()


def double_sha256(data):
    """A standard compound hash."""
    return bytes_as_revhex(hashlib.sha256(hashlib.sha256(data).digest()).digest())


def hash160(data):
    """A standard compound hash."""
    return ripemd160(hashlib.sha256(data).digest()).digest()


"""
The MIT License (MIT)

Copyright (c) 2013 by Richard Kiss

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""



from ..encoding.sec import is_sec, public_pair_to_hash160_sec, sec_to_public_pair, EncodingError

from pycoin.coins.SolutionChecker import ScriptError
from pycoin.intbytes import byte2int
from pycoin.satoshi.checksigops import parse_signature_blob
from pycoin.satoshi.der import UnexpectedDER


class WhoSigned(object):
    def __init__(self, script_tools, address_api, generator):
        self._script_tools = script_tools
        self._address = address_api
        self._generator = generator
        for _ in "CHECKSIG CHECKSIGVERIFY CHECKMULTISIG CHECKMULTISIGVERIFY".split():
            setattr(self, "OP_%s" % _, byte2int(self._script_tools.compile('OP_%s' % _)))

    def solution_blobs(self, tx, tx_in_idx):
        """
        This iterator yields data blobs that appear in the last solution_script or the witness.
        """
        sc = tx.SolutionChecker(tx)
        tx_context = sc.tx_context_for_idx(tx_in_idx)
        # set solution_stack in case there are no results from puzzle_and_solution_iterator
        solution_stack = []
        for puzzle_script, solution_stack, flags, sighash_f in sc.puzzle_and_solution_iterator(tx_context):
            pass
            # we only care about the last one
        for s in solution_stack:
            yield s

    def _handle_checksig(self, vmc):
        s = list(vmc.stack)
        sec_blob = sig_blob = None
        sig_hash = 0
        try:
            sec_blob = vmc.pop()
            sig_blob = vmc.pop()
            try:
                sig_pair, sig_type = parse_signature_blob(sig_blob)
                sig_hash = vmc.signature_for_hash_type_f(sig_type, [sig_blob], vmc)
            except ValueError:
                pass
        except (IndexError, ValueError):
            pass

        return ([sec_blob], [(sig_blob, sig_hash)])

    def _handle_checkmultisig(self, vmc):
        sec_blobs = []
        sig_blobs = []
        s = list(vmc.stack)
        try:
            key_count = vmc.pop_int()
            while key_count > 0:
                key_count -= 1
                sec_blobs.append(vmc.pop())

            signature_count = vmc.pop_int()
            while signature_count > 0:
                signature_count -= 1
                sig_blob = vmc.pop()
                sig_hash = 0
                try:
                    sig_pair, sig_type = parse_signature_blob(sig_blob)
                    sig_hash = vmc.signature_for_hash_type_f(sig_type, [sig_blob], vmc)
                except ValueError:
                    pass
                sig_blobs.append((sig_blob, sig_hash))
        except IndexError:
            pass
        vmc.stack = s
        return (sec_blobs, sig_blobs)

    def extract_secs(self, tx, tx_in_idx):
        for sec_blobs, sig_and_hash_pairs in self.extract_secs_and_signatures(tx, tx_in_idx):
            for blob in sec_blobs:
                yield blob

    def extract_signatures(self, tx, tx_in_idx):
        for sec_blobs, sig_and_hash_pairs in self.extract_secs_and_signatures(tx, tx_in_idx):
            for sig_blob, sig_hash in sig_and_hash_pairs:
                try:
                    sig_pair, sig_type = parse_signature_blob(sig_blob)
                    yield (sig_blob, sig_hash)
                except (ValueError, TypeError, UnexpectedDER, ScriptError):
                    continue

    def extract_secs_and_signatures(self, tx, tx_in_idx):
        """
        List[Tuple[List[bytes], List[Tuple[bytes, SigHashFType]]]
        """
        sc = tx.SolutionChecker(tx)
        tx_context = sc.tx_context_for_idx(tx_in_idx)

        blobs_for_sig_ops = []
        def traceback_f(opcode, data, pc, vmc):
            if opcode in (self.OP_CHECKSIG, self.OP_CHECKSIGVERIFY):
                blobs_for_sig_ops.append(self._handle_checksig(vmc))
            if opcode in (self.OP_CHECKMULTISIG, self.OP_CHECKMULTISIGVERIFY):
                blobs_for_sig_ops.append(self._handle_checkmultisig(vmc))
            return

        try:
            sc.check_solution(tx_context, traceback_f=traceback_f)
        except ScriptError:
            pass

        return blobs_for_sig_ops

    def public_pairs_for_script(self, tx, tx_in_idx, generator):
        """
        For a given script, iterate over and pull out public pairs encoded as sec values.
        """
        public_pairs = []
        for sec in self.extract_secs(tx, tx_in_idx):
            try:
                public_pairs.append(sec_to_public_pair(sec, generator))
            except EncodingError:
                pass
        return public_pairs

    def public_pairs_signed(self, tx, tx_in_idx):
        signed_by = []

        public_pairs = self.public_pairs_for_script(tx, tx_in_idx, self._generator)

        for signature, sig_hash in self.extract_signatures(tx, tx_in_idx):
            sig_pair, sig_type = parse_signature_blob(signature)

            for public_pair in public_pairs:
                if self._generator.verify(public_pair, sig_hash, sig_pair):
                    signed_by.append((public_pair, sig_pair, sig_type))
        return signed_by

    def who_signed_tx(self, tx, tx_in_idx):
        """
        Given a transaction (tx) an input index (tx_in_idx), attempt to figure
        out which addresses where used in signing (so far). This method
        depends on tx.unspents being properly configured. This should work on
        partially-signed MULTISIG transactions (it will return as many
        addresses as there are good signatures).
        Returns a list of (public_pairs, sig_type) pairs.
        """
        public_pair_sig_type_list = self.public_pairs_signed(tx, tx_in_idx)
        sig_type_list = [pp[-1] for pp in public_pair_sig_type_list]
        hash160_list = [public_pair_to_hash160_sec(pp[0]) for pp in public_pair_sig_type_list]
        address_list = [self._address.for_p2pkh(h160) for h160 in hash160_list]
        return list(zip(address_list, sig_type_list))



import hashlib
import hmac
import struct

from ..encoding.bytes32 import from_bytes_32
from ..encoding.sec import sec_to_public_pair
from .Key import Key


class HDSeed:
    def __init__(self, data):
        self.data = data

    @classmethod
    def deserialize(class_, data):
        parent_fingerprint, child_index = struct.unpack(">4sL", data[5:13])
        d = dict(chain_code=data[13:45], depth=ord(data[4:5]), parent_fingerprint=parent_fingerprint,
                 child_index=child_index)
        is_private = (data[45:46] == b'\0')
        if is_private:
            d["secret_exponent"] = from_bytes_32(data[46:])
        else:
            d["public_pair"] = sec_to_public_pair(data[45:], generator=class_._generator)
        return class_(**d)

    def __repr__(self):
        return "<HDSeed with hash %s>" % hashlib.sha256(self.data).hexdigest()[:8]



from pycoin.block import Block
from pycoin.coins.bitcoin.ScriptTools import BitcoinScriptTools
from pycoin.coins.bitcoin.Tx import Tx
from pycoin.coins.exceptions import ValidationFailureError
from pycoin.coins.tx_utils import create_tx, split_with_remainder, distribute_from_split_pool, sign_tx, create_signed_tx
from pycoin.coins.SolutionChecker import ScriptError
from pycoin.contrib.msg_signing import MessageSigner
from pycoin.contrib.who_signed import WhoSigned
from pycoin.ecdsa.secp256k1 import secp256k1_generator
from pycoin.encoding.b58 import b2a_hashed_base58
from pycoin.key.HDSeed import HDSeed
from pycoin.key.Keychain import Keychain
from pycoin.key.Key import Key, InvalidSecretExponentError, InvalidPublicPairError
from pycoin.key.BIP32Node import BIP32Node
from pycoin.key.BIP49Node import BIP49Node
from pycoin.key.BIP84Node import BIP84Node
from pycoin.key.electrum import ElectrumWallet
from pycoin.message.make_parser_and_packer import (
    make_parser_and_packer, standard_messages,
    standard_message_post_unpacks, standard_streamer, standard_parsing_functions
)
from pycoin.encoding.hexbytes import b2h, h2b
from pycoin.satoshi import errno, flags
from pycoin.solve.utils import build_hash160_lookup, build_p2sh_lookup, build_sec_lookup
from pycoin.vm.annotate import Annotate

from .AddressAPI import make_address_api
from .ParseAPI import ParseAPI
from .ContractAPI import ContractAPI
from .parseable_str import parseable_str


class API(object):
    pass


class Network(object):
    def __init__(self, symbol, network_name, subnet_name):
        self.symbol = symbol
        self.network_name = network_name
        self.subnet_name = subnet_name

    def full_name(self):
        return "%s %s" % (self.network_name, self.subnet_name)

    def __repr__(self):
        return "<Network %s>" % self.full_name()


def make_output_for_secret_exponent(Key):
    def f(secret_exponent):
        yield ("secret_exponent", '%d' % secret_exponent, None)
        yield ("secret_exponent_hex", '%x' % secret_exponent, " hex")
        key = Key(secret_exponent)
        yield ("wif", key.wif(is_compressed=True), None)
        yield ("wif_uncompressed", key.wif(is_compressed=False), " uncompressed")
    return f


def make_output_for_public_pair(Key, network):
    def f(public_pair):
        yield ("public_pair_x", '%d' % public_pair[0], None)
        yield ("public_pair_y", '%d' % public_pair[1], None)
        yield ("public_pair_x_hex", '%x' % public_pair[0], " x as hex")
        yield ("public_pair_y_hex", '%x' % public_pair[1], " y as hex")
        yield ("y_parity", "odd" if (public_pair[1] & 1) else "even", None)

        key = Key(public_pair=public_pair)
        yield ("key_pair_as_sec", b2h(key.sec(is_compressed=True)), None)
        yield ("key_pair_as_sec_uncompressed", b2h(key.sec(is_compressed=False)), " uncompressed")

        network_name = network.network_name
        hash160_u = key.hash160(is_compressed=False)
        hash160_c = key.hash160(is_compressed=True)

        yield ("hash160", b2h(hash160_c), None)

        if hash160_c and hash160_u:
            yield ("hash160_uncompressed", b2h(hash160_u), " uncompressed")

        address = network.address.for_p2pkh(hash160_c)
        yield ("address", address, "%s address" % network_name)
        yield ("%s_address" % network.symbol, address, "legacy")

        address = key.address(is_compressed=False)
        yield ("address_uncompressed", address, "%s address uncompressed" % network_name)
        yield ("%s_address_uncompressed" % network.symbol, address, "legacy")

        # don't print segwit addresses unless we're sure we have a compressed key
        if hash160_c and hasattr(network.address, "for_p2pkh_wit"):
            address_segwit = network.address.for_p2pkh_wit(hash160_c)
            if address_segwit:
                # this network seems to support segwit
                yield ("address_segwit", address_segwit, "%s segwit address" % network_name)
                yield ("%s_address_segwit" % network.symbol, address_segwit, "legacy")

                p2sh_script = network.contract.for_p2pkh_wit(hash160_c)
                p2s_address = network.address.for_p2s(p2sh_script)
                if p2s_address:
                    yield ("p2sh_segwit", p2s_address, None)

                p2sh_script_hex = b2h(p2sh_script)
                yield ("p2sh_segwit_script", p2sh_script_hex, " corresponding p2sh script")

    return f


def create_bitcoinish_network(symbol, network_name, subnet_name, **kwargs):
    # potential kwargs:
    #   tx, block, magic_header_hex, default_port, dns_bootstrap,
    #   wif_prefix_hex, address_prefix_hex, pay_to_script_prefix_hex
    #   bip32_prv_prefix_hex, bip32_pub_prefix_hex, sec_prefix, script_tools
    #   bip49_prv_prefix, bip49_pub_prefix, bip84_prv_prefix, bip84_pub_prefix

    network = Network(symbol, network_name, subnet_name)

    generator = kwargs.get("generator", secp256k1_generator)
    kwargs.setdefault("sec_prefix", "%sSEC" % symbol.upper())
    KEYS_TO_H2B = ("bip32_prv_prefix bip32_pub_prefix bip49_prv_prefix bip49_pub_prefix "
                   "bip84_prv_prefix bip84_pub_prefix wif_prefix address_prefix "
                   "pay_to_script_prefix sec_prefix magic_header").split()
    for k in KEYS_TO_H2B:
        k_hex = "%s_hex" % k
        if k_hex in kwargs:
            kwargs[k] = h2b(kwargs[k_hex])

    script_tools = kwargs.get("script_tools", BitcoinScriptTools)

    UI_KEYS = ("bip32_prv_prefix bip32_pub_prefix bip49_prv_prefix bip49_pub_prefix "
               "bip84_prv_prefix bip84_pub_prefix wif_prefix sec_prefix "
               "address_prefix pay_to_script_prefix bech32_hrp").split()
    ui_kwargs = {k: kwargs[k] for k in UI_KEYS if k in kwargs}

    _bip32_prv_prefix = ui_kwargs.get("bip32_prv_prefix")
    _bip32_pub_prefix = ui_kwargs.get("bip32_pub_prefix")
    _wif_prefix = ui_kwargs.get("wif_prefix")
    _sec_prefix = ui_kwargs.get("sec_prefix")

    def bip32_as_string(blob, as_private):
        prefix = ui_kwargs.get("bip32_%s_prefix" % ("prv" if as_private else "pub"))
        return b2a_hashed_base58(prefix + blob)

    def bip49_as_string(blob, as_private):
        prefix = ui_kwargs.get("bip49_%s_prefix" % ("prv" if as_private else "pub"))
        return b2a_hashed_base58(prefix + blob)

    def bip84_as_string(blob, as_private):
        prefix = ui_kwargs.get("bip84_%s_prefix" % ("prv" if as_private else "pub"))
        return b2a_hashed_base58(prefix + blob)

    def wif_for_blob(blob):
        return b2a_hashed_base58(_wif_prefix + blob)

    def sec_text_for_blob(blob):
        return _sec_prefix + b2h(blob)

    NetworkKey = Key.make_subclass(symbol, network=network, generator=generator)
    NetworkElectrumKey = ElectrumWallet.make_subclass(symbol, network=network, generator=generator)
    NetworkBIP32Node = BIP32Node.make_subclass(symbol, network=network, generator=generator)
    NetworkBIP49Node = BIP49Node.make_subclass(symbol, network=network, generator=generator)
    NetworkBIP84Node = BIP84Node.make_subclass(symbol, network=network, generator=generator)

    NETWORK_KEYS = "network_name subnet_name dns_bootstrap default_port magic_header".split()
    for k in NETWORK_KEYS:
        if k in kwargs:
            setattr(network, k, kwargs[k])

    network.Tx = network.tx = kwargs.get("tx") or Tx
    network.Block = network.block = kwargs.get("block") or Block.make_subclass(symbol, network.tx)

    streamer = standard_streamer(standard_parsing_functions(network.block, network.tx))

    network.message = API()
    network.message.parse, network.message.pack = make_parser_and_packer(
        streamer, standard_messages(), standard_message_post_unpacks(streamer))

    network.output_for_secret_exponent = make_output_for_secret_exponent(NetworkKey)
    network.output_for_public_pair = make_output_for_public_pair(NetworkKey, network)

    network.keychain = Keychain

    parse_api_class = kwargs.get("parse_api_class", ParseAPI)
    network.parse = parse_api_class(network, **ui_kwargs)

    network.contract = ContractAPI(network, script_tools)

    network.address = make_address_api(network.contract, **ui_kwargs)

    def keys_private(secret_exponent, is_compressed=True):
        return NetworkKey(secret_exponent=secret_exponent, is_compressed=is_compressed)

    def keys_public(item, is_compressed=None):
        if isinstance(item, tuple):
            if is_compressed is None:
                is_compressed = True
            # it's a public pair
            return NetworkKey(public_pair=item, is_compressed=is_compressed)
        if is_compressed is not None:
            raise ValueError("can't set is_compressed from sec")
        return NetworkKey.from_sec(item)

    network.keys = API()
    network.keys.private = keys_private
    network.keys.public = keys_public

    def electrum_seed(seed):
        return NetworkElectrumKey(initial_key=seed)

    def electrum_private(master_private_key):
        return NetworkElectrumKey(master_private_key=master_private_key)

    def electrum_public(master_public_key):
        return NetworkElectrumKey(master_public_key=master_public_key)

    network.keys.bip32_seed = NetworkBIP32Node.from_master_secret
    network.keys.bip32_deserialize = NetworkBIP32Node.deserialize
    network.keys.bip49_deserialize = NetworkBIP49Node.deserialize
    network.keys.bip84_deserialize = NetworkBIP84Node.deserialize

    network.keys.electrum_seed = electrum_seed
    network.keys.electrum_private = electrum_private
    network.keys.electrum_public = electrum_public
    network.keys.InvalidSecretExponentError = InvalidSecretExponentError
    network.keys.InvalidPublicPairError = InvalidPublicPairError

    network.msg = API()
    message_signer = MessageSigner(network, generator)
    network.msg.sign = message_signer.sign_message
    network.msg.verify = message_signer.verify_message
    network.msg.parse_signed = message_signer.parse_signed_message
    network.msg.hash_for_signing = message_signer.hash_for_signing
    network.msg.signature_for_message_hash = message_signer.signature_for_message_hash
    network.msg.pair_for_message_hash = message_signer.pair_for_message_hash
    network.script = script_tools

    network.bip32_as_string = bip32_as_string
    network.bip49_as_string = bip49_as_string
    network.bip84_as_string = bip84_as_string
    network.sec_text_for_blob = sec_text_for_blob
    network.wif_for_blob = wif_for_blob

    def network_build_hash160_lookup(iter):
        return build_hash160_lookup(iter, [generator])

    network.tx.solve = API()
    network.tx.solve.build_hash160_lookup = network_build_hash160_lookup
    network.tx.solve.build_p2sh_lookup = build_p2sh_lookup
    network.tx.solve.build_sec_lookup = build_sec_lookup

    network.validator = API()
    network.validator.ScriptError = ScriptError
    network.validator.ValidationFailureError = ValidationFailureError
    network.validator.errno = errno
    network.validator.flags = flags

    def my_create_tx(*args, **kwargs):
        return create_tx(network, *args, **kwargs)

    def my_sign_tx(*args, **kwargs):
        return sign_tx(network, *args, **kwargs)

    def my_create_signed_tx(*args, **kwargs):
        return create_signed_tx(network, *args, **kwargs)

    def my_split_with_remainder(*args, **kwargs):
        return split_with_remainder(network, *args, **kwargs)

    network.tx_utils = API()
    network.tx_utils.create_tx = my_create_tx
    network.tx_utils.sign_tx = my_sign_tx
    network.tx_utils.create_signed_tx = my_create_signed_tx
    network.tx_utils.split_with_remainder = my_split_with_remainder
    network.tx_utils.distribute_from_split_pool = distribute_from_split_pool

    network.annotate = Annotate(script_tools, network.address)

    network.who_signed = WhoSigned(script_tools, network.address, generator)

    network.str = parseable_str

    network.generator = generator

    return network



from pycoin.block import Block
from pycoin.coins.bitcoin.Tx import Tx
from pycoin.satoshi.satoshi_int import parse_satoshi_int
from pycoin.satoshi.satoshi_string import parse_satoshi_string
from pycoin.satoshi.satoshi_struct import parse_struct


class LTCTx(Tx):
    @classmethod
    def parse(class_, f):
        """Parse a Bitcoin transaction Tx.
        :param f: a file-like object that contains a binary streamed transaction
        """
        (version,) = parse_struct("L", f)
        v1 = ord(f.read(1))
        is_segwit = v1 == 0
        has_mweb = False
        if is_segwit:
            flag = ord(f.read(1))
            if flag == 0:
                raise ValueError("bad flag in segwit")
            has_mweb = flag & 0x8 != 0
            is_segwit = flag & 0x1 != 0
            v1 = None
        count = parse_satoshi_int(f, v=v1)
        txs_in = []
        for i in range(count):
            txs_in.append(class_.TxIn.parse(f))
        count = parse_satoshi_int(f)
        txs_out = []
        for i in range(count):
            txs_out.append(class_.TxOut.parse(f))
        if is_segwit:
            for tx_in in txs_in:
                stack = []
                count = parse_satoshi_int(f)
                for i in range(count):
                    stack.append(parse_satoshi_string(f))
                tx_in.witness = stack
        if has_mweb:
            mweb_tx_type = ord(f.read(1))
            if mweb_tx_type:
                # TODO: read serialized MWEB transaction data
                # TODO: save this away
                pass
        (lock_time,) = parse_struct("L", f)
        return class_(version, txs_in, txs_out, lock_time)


class LTCBlock(Block):
    Tx = LTCTx



from pycoin.networks.bitcoinish import create_bitcoinish_network

network = create_bitcoinish_network(
    symbol="XMY", network_name="Myriadcoin", subnet_name="mainnet",
    wif_prefix_hex="b2", address_prefix_hex="32", pay_to_script_prefix_hex="09",
    bip32_prv_prefix_hex="0488ade4", bip32_pub_prefix_hex="0488b21e")
